
import typing

from ._pathfinder import (
    Segment,
    
    pathfinder_modify_swerve,
    pathfinder_modify_tank,
)

__all__ = [
    'SwerveModifier',
    'TankModifier',
]


class SwerveModifier:
    """
        The Swerve Modifier will take in a Source Trajectory and spit out 4 trajectories, 1 for each wheel on the drive.
        This is commonly used in robotics for robots with 4 individual wheels in a 'swerve' configuration, where each wheel
        can rotate to a specified heading while still being powered.
        
        The Source Trajectory is measured from the centre of the drive base. The modification will not modify the central
        trajectory
    """
    
    def __init__(self, source: typing.List[Segment]) -> None:
        """Create an instance of the modifier
        :param source: The source (center) trajectory
        """
        self.source = source
        self.fl = None
        self.fr = None
        self.bl = None
        self.br = None
    
    def modify(self, wheelbase_width: float, wheelbase_depth: float) -> 'SwerveModifier':
        """Generate the Trajectory Modification
        
        :param wheelbase_width: The width (in meters) between the individual left-right sides of the drivebase
        :param wheelbase_depth: The width (in meters) between the individual front-back sides of the drivebase
        :returns:                  self
        """
        self.fl, self.fr, self.bl, self.br = pathfinder_modify_swerve(self.source, wheelbase_width, wheelbase_depth)
        return self
    
    def getSourceTrajectory(self) -> typing.List[Segment]:
        """Get the initial source trajectory"""
        return self.source
    
    def getFrontLeftTrajectory(self) -> typing.List[Segment]:
        """Get the trajectory for the front-left wheel of the drive base"""
        return self.fl
    
    def getFrontRightTrajectory(self) -> typing.List[Segment]:
        """Get the trajectory for the front-right wheel of the drive base"""
        return self.fr
    
    def getBackLeftTrajectory(self) -> typing.List[Segment]:
        """Get the trajectory for the back-left wheel of the drive base"""
        return self.bl
    
    def getBackRightTrajectory(self) -> typing.List[Segment]:
        """Get the trajectory for the back-right wheel of the drive base"""
        return self.br


class TankModifier:
    """
        The Tank Modifier will take in a Source Trajectory and a Wheelbase Width and spit out a Trajectory for each
        side of the wheelbase. This is commonly used in robotics for robots which have a drive system similar
        to a 'tank', where individual parallel sides are driven independently
        
        The Source Trajectory is measured from the centre of the drive base. The modification will not modify the central
        trajectory
    """
    
    def __init__(self, source: typing.List[Segment]) -> None:
        """Create an instance of the modifier
        :param source: The source (center) trajectory
        """
        self.source = source
        self.left = None
        self.right = None
    
    def modify(self, wheelbase_width: float) -> 'TankModifier':
        """Generate the Trajectory Modification
        
        :param wheelbase_width: The width (in meters) between the individual sides of the drivebase
        :returns:                  self
        """
        self.left, self.right = pathfinder_modify_tank(self.source, wheelbase_width)
        return self
    
    def getSourceTrajectory(self) -> typing.List[Segment]:
        """Get the initial source trajectory"""
        return self.source
    
    def getLeftTrajectory(self) -> typing.List[Segment]:
        """Get the trajectory for the left side of the drive base"""
        return self.left
    
    def getRightTrajectory(self) -> typing.List[Segment]:
        """Get the trajectory for the right side of the drive base"""
        return self.right




import math
import typing

from ._pathfinder import (
    DistanceFollower as _DistanceFollower,
    EncoderFollower as _EncoderFollower,
    
    EncoderConfig,
    FollowerConfig,
    Segment,
    
    pathfinder_follow_distance2,
    pathfinder_follow_encoder2,
)

__all__ = [
    'DistanceFollower',
    'EncoderFollower',
]


class DistanceFollower(_DistanceFollower):
    """
        The DistanceFollower is an object designed to follow a trajectory based on distance covered input. This class can be used
        for Tank or Swerve drive implementations.
    """
    
    def __init__(self, trajectory: typing.List[Segment]):
        super().__init__()
        self.trajectory = trajectory
        self.cfg = FollowerConfig()
    
    def setTrajectory(self, trajectory: typing.List[Segment]) -> None:
        """Set a new trajectory to follow, and reset the cumulative errors and segment counts"""
        self.trajectory = trajectory
        self.reset()
    
    def configurePIDVA(self, kp: float, ki: float, kd: float, kv: float, ka: float) -> None:
        """Configure the PID/VA Variables for the Follower
        
        :param kp: The proportional term. This is usually quite high (0.8 - 1.0 are common values)
        :param ki: The integral term. Currently unused.
        :param kd: The derivative term. Adjust this if you are unhappy with the tracking of the follower. 0.0 is the default
        :param kv: The velocity ratio. This should be 1 over your maximum velocity @ 100% throttle.
                   This converts m/s given by the algorithm to a scale of -1..1 to be used by your
                   motor controllers
        :param ka: The acceleration term. Adjust this if you want to reach higher or lower speeds faster. 0.0 is the default
        """
        self.cfg.kp = kp
        self.cfg.ki = ki
        self.cfg.kd = kd
        self.cfg.kv = kv
        self.cfg.ka = ka
    
    def reset(self) -> None:
        """Reset the follower to start again. Encoders must be reconfigured."""
        self.last_error = 0
        self.segment = 0
    
    def calculate(self, distance_covered: float) -> float:
        """Calculate the desired output for the motors, based on the distance the robot has covered.
        This does not account for heading of the robot. To account for heading, add some extra terms in your control
        loop for realignment based on gyroscope input and the desired heading given by this object.
        
        :param distance_covered: The distance covered in meters
        :returns:                  The desired output for your motor controller
        """
        tlen = len(self.trajectory)
        if self.segment >= tlen:
            self.finished = 1
            self.output = 0
            self.heading = self.trajectory[-1].heading
            return 0.0
        else:
            return pathfinder_follow_distance2(self.cfg, self, self.trajectory[self.segment], tlen, distance_covered)
    
    def getHeading(self) -> float:
        """:returns: the desired heading of the current point in the trajectory"""
        return self.heading
    
    def getSegment(self) -> Segment:
        """:returns: the current segment being operated on"""
        return self.trajectory[self.segment]
    
    def isFinished(self) -> bool:
        """:returns: whether we have finished tracking this trajectory or not."""
        return self.segment >= len(self.trajectory)


class EncoderFollower(_EncoderFollower):
    """
        The EncoderFollower is an object designed to follow a trajectory based on encoder input. This class can be used
        for Tank or Swerve drive implementations.
    """
    
    def __init__(self, trajectory: typing.List[Segment]):
        super().__init__()
        self.trajectory = trajectory
        self.cfg = EncoderConfig()
    
    def setTrajectory(self, trajectory: typing.List[Segment]) -> None:
        """Set a new trajectory to follow, and reset the cumulative errors and segment counts"""
        self.trajectory = trajectory
        self.reset()
    
    def configurePIDVA(self, kp: float, ki: float, kd: float, kv: float, ka: float) -> None:
        """Configure the PID/VA Variables for the Follower
        
        :param kp: The proportional term. This is usually quite high (0.8 - 1.0 are common values)
        :param ki: The integral term. Currently unused.
        :param kd: The derivative term. Adjust this if you are unhappy with the tracking of the follower. 0.0 is the default
        :param kv: The velocity ratio. This should be 1 over your maximum velocity @ 100% throttle.
                   This converts m/s given by the algorithm to a scale of -1..1 to be used by your
                   motor controllers
        :param ka: The acceleration term. Adjust this if you want to reach higher or lower speeds faster. 0.0 is the default
        """
        self.cfg.kp = kp
        self.cfg.ki = ki
        self.cfg.kd = kd
        self.cfg.kv = kv
        self.cfg.ka = ka
    
    def configureEncoder(self, initial_position: int, ticks_per_revolution: int, wheel_diameter: float) -> None:
        """Configure the Encoders being used in the follower.
        
        :param initial_position: The initial 'offset' of your encoder. This should be set to the encoder value just
                                 before you start to track
        :param ticks_per_revolution: How many ticks per revolution the encoder has
        :param wheel_diameter: The diameter of your wheels (or pulleys for track systems) in meters
        """
        self.cfg.initial_position = initial_position
        self.cfg.ticks_per_revolution = ticks_per_revolution
        self.cfg.wheel_circumference = math.pi * wheel_diameter
    
    def reset(self) -> None:
        """Reset the follower to start again. Encoders must be reconfigured."""
        self.last_error = 0
        self.segment = 0
    
    def calculate(self, encoder_tick: int) -> float:
        """Calculate the desired output for the motors, based on the amount of ticks the encoder has gone through.
        This does not account for heading of the robot. To account for heading, add some extra terms in your control
        loop for realignment based on gyroscope input and the desired heading given by this object.
        
        :param encoder_tick: The amount of ticks the encoder has currently measured.
        :returns:             The desired output for your motor controller
        """
        tlen = len(self.trajectory)
        if self.segment >= tlen:
            self.finished = 1
            self.output = 0
            self.heading = self.trajectory[-1].heading
            return 0.0
        else:
            return pathfinder_follow_encoder2(self.cfg, self, self.trajectory[self.segment], tlen, encoder_tick)

    def getHeading(self) -> float:
        """:returns: the desired heading of the current point in the trajectory"""
        return self.heading
    
    def getSegment(self) -> Segment:
        """:returns: the current segment being operated on"""
        return self.trajectory[self.segment]
    
    def isFinished(self) -> bool:
        """:returns: whether we have finished tracking this trajectory or not."""
        return self.segment >= len(self.trajectory)




from ._pathfinder import (
    __version__,
    
    Segment,
    TrajectoryInfo,
    Waypoint,
    
    pathfinder_generate as generate,
    
    pathfinder_deserialize as deserialize,
    pathfinder_serialize as serialize,
    pathfinder_serialize_csv as serialize_csv,
    
    pf_fit_hermite_pre as FIT_HERMITE_PRE,
    pf_fit_hermite_cubic as FIT_HERMITE_CUBIC,
    pf_fit_hermite_quintic as FIT_HERMITE_QUINTIC,
    
    SAMPLES_FAST,
    SAMPLES_LOW,
    SAMPLES_HIGH,
)

from . import (
    followers,
    modifiers,
)

# compability with Pathfinder-Java
import math

r2d = math.degrees
d2r = math.radians

def boundHalfDegrees(degrees):
    '''Bound an angle (in degrees) to -180 to 180 degrees.'''
    degrees = math.fmod(degrees, 360.0)
    if degrees >= 180.0:
        degrees = degrees - 360.0
    elif degrees <= -180.0:
        degrees = degrees + 360.0
    return degrees

def deserialize_csv(fname):
    '''Read a Trajectory from a CSV File'''
    import csv
    with open(fname, 'r') as fp:
        csviter =  csv.reader(fp)
        next(csviter)
        return [Segment(*map(float, row)) for row in csviter]




#
# Much of this copied from https://github.com/pybind/python_example.git
#

import sys
if sys.version_info[0] < 3:
    sys.stderr.write("ERROR: robotpy-pathfinder requires python 3!")
    exit(1)

import fnmatch
import os
from os.path import dirname, exists, join
from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext
import subprocess
import sys
import setuptools

setup_dir = dirname(__file__)
git_dir = join(setup_dir, '.git')
base_package = 'pathfinder'
version_file = join(setup_dir, base_package, 'version.py')

# Automatically generate a version.py based on the git version
if exists(git_dir):
    p = subprocess.Popen(["git", "describe", "--tags", "--long", "--dirty=-dirty"],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    # Make sure the git version has at least one tag
    if err:
        print("Error: You need to create a tag for this repo to use the builder")
        sys.exit(1)

    # Convert git version to PEP440 compliant version
    # - Older versions of pip choke on local identifiers, so we can't include the git commit
    v, commits, local = out.decode('utf-8').rstrip().split('-', 2)
    if commits != '0' or '-dirty' in local:
        v = '%s.post0.dev%s' % (v, commits)

    # Create the version.py file
    with open(version_file, 'w') as fp:
        fp.write("# Autogenerated by setup.py\n__version__ = '{0}'".format(v))

if exists(version_file):
    with open(join(setup_dir, base_package, 'version.py'), 'r') as fp:
        exec(fp.read(), globals())
else:
    __version__ = 'master'

with open(join(setup_dir, 'README.rst'), 'r') as readme_file:
    long_description = readme_file.read()


#
# pybind-specific compilation stuff
#

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is preferred over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            # TODO: this feels like a hack
            if not os.environ.get('RPY_DEBUG'):
                opts.append('-s') # strip
                opts.append('-g0') # remove debug symbols
            else:
                opts.append('-O0')
            c11arg = cpp_flag(self.compiler)
            opts.append(c11arg)
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
            
            # HACK: clang doesn't allow compiling .c files with the -std=c++11
            #       flag... so remove it
            if sys.platform == 'darwin':
                _orig_compile = self.compiler._compile
                def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
                    if src.endswith('.c'):
                        extra_postargs = extra_postargs[:]
                        extra_postargs.remove(c11arg)
                    return _orig_compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
                self.compiler._compile = _compile
                
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)

def recursive_glob(d):
    for root, _, filenames in os.walk(d):
        for fname in fnmatch.filter(filenames, '*.c'):
            yield join(root, fname)

ext_modules = [
    Extension(
        'pathfinder._pathfinder',
        [
            'pathfinder/_pathfinder.cpp',
        ] + \
            list(recursive_glob('pathfinder_src/Pathfinder-Core/src')),
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            'pathfinder_src/Pathfinder-Core/include',
        ],
        language='c++',
    ),
]

setup(
    name='robotpy-pathfinder',
    version=__version__,
    author='Dustin Spicuzza',
    author_email='dustin@virtualroadside.com',
    url='https://github.com/robotpy/robotpy-pathfinder',
    description="RobotPy bindings for Jaci R's PathFinder library",
    long_description=long_description,
    packages=find_packages(),
    ext_modules=ext_modules,
    setup_requires=['pybind11>=2.2'],
    license="MIT",
    zip_safe=False,
    cmdclass={'build_ext': BuildExt},
)
